import logging
from typing import List, Optional

from prance import ResolvingParser

from caffoa.converter import to_camelcase, parse_type, is_primitive
from caffoa.model import EndPoint, Parameter, Response
from caffoa.object_parser import BaseObjectParser


class PathParser:
    def __init__(self, parser: ResolvingParser):
        self.suffix = ""
        self.prefix = ""
        self.parser = parser
        self.responses = dict()
        self.bodies = dict()
        self.known_types = list()

    def create_returns(self, model_parser: ResolvingParser):
        for path, options in model_parser.specification['paths'].items():
            for operation, config in options.items():
                operation = operation.lower()
                if operation in ["parameters"]:
                    continue
                if "operationId" not in config:
                    raise Warning(f"operationId is missing for {path}: {operation}")
                operation_id = config["operationId"]
                code = None
                try:
                    responses = list()
                    for code, data in config.get("responses").items():
                        if "$ref" in data:
                            ref = data["$ref"].split('/')[-1]
                            data = model_parser.specification['components']['responses'][ref]
                        response = Response(code)
                        response.content = self.parse_content(data)
                        responses.append(response)
                    self.responses[operation_id] = responses

                except Warning as msg:
                    logging.warning(
                        f"Cannot generate typed responses for {msg} for {operation_id} ({path}): {operation}/{code}")

    def create_bodies(self, model_parser: ResolvingParser):
        for path, options in model_parser.specification['paths'].items():
            for operation, config in options.items():
                operation = operation.lower()
                if operation in ["parameters"]:
                    continue
                if "operationId" not in config:
                    raise Warning(f"operationId is missing for {path}: {operation}")
                operation_id = config["operationId"]
                if "requestBody" in config:
                    try:
                        self.bodies[operation_id] = self.parse_content(config["requestBody"])
                    except Warning as msg:
                        logging.warning(
                            f"Cannot generate typed requestBody for {msg} for {operation_id} ({path}): {operation}")

    def parse_content(self, config: dict) -> Optional[str]:

        if "content" in config:
            content = config["content"]
            keys = list(content.keys())
            if len(keys) > 1:
                raise Warning(
                    f"multiple possible responses")
            type = str(keys[0])
            if type.lower() != "application/json":
                raise Warning(
                    f"type {type}. Only application/json is currently supported for")
            try:
                schema = content[type]['schema']
            except KeyError:
                raise Warning(
                    f"type {type}. no schema found")
            if "$ref" in schema:
                return self.name_for_ref(schema["$ref"])
            if "type" in schema and is_primitive(schema["type"]):
                return parse_type(schema)
            if "type" in schema and schema["type"].lower() == "array":
                if "$ref" in schema["items"]:
                    name = self.name_for_ref(schema["items"]["$ref"])
                    return f"IEnumerable<{name}>"
                if "type" in schema["items"] and is_primitive(schema["items"]["type"]):
                    type = parse_type(schema["items"])
                    return f"IEnumerable<{type}>"
                raise Warning("array with complex type that is not a reference")

            raise Warning(
                "complex type. Only array, ref or basic types are supported.")

        return None

    def parse(self) -> List[EndPoint]:
        endpoints = list()
        for path, options in self.parser.specification['paths'].items():
            path = str(path).strip('/')
            base_parameters = list()
            if "parameters" in options:
                base_parameters = self.parse_params(options["parameters"])
            for operation, config in options.items():
                operation = operation.lower()
                if operation in ["parameters"]:
                    continue
                if "operationId" not in config:
                    raise Warning(f"OperationId is missing for {path}: {operation}")
                operation_id = config['operationId']
                operation_name = self.operation_name(operation_id)
                documentation = config['description'].split("\n")
                for response, response_data in config["responses"].items():
                    documentation.append(f"{response} -> {response_data['description']}")

                parameters = base_parameters.copy()
                if "parameters" in config:
                    parameters.extend(self.parse_params(config["parameters"]))
                needs_content = "requestBody" in config
                ep = EndPoint(operation_name, path, operation, parameters, documentation, needs_content)
                ep.responses = self.responses.get(operation_id)
                ep.body = self.bodies.get(operation_id)
                endpoints.append(ep)
        return endpoints

    def name_for_ref(self, ref):
        name = BaseObjectParser(self.prefix, self.suffix).class_name_from_ref(ref)
        if name in self.known_types:
            return dict(self.known_types[name])["type"]
        return name

    @staticmethod
    def operation_name(operation_id) -> str:
        return to_camelcase(operation_id) + "Async"

    @staticmethod
    def parse_params(params: list) -> list:
        parameters = list()
        for param in params:
            if param['in'].lower() == "path":
                parameters.append(Parameter(param['name'], parse_type(param['schema']), param.get('description')))
        return parameters



import logging
import os
from typing import List

from caffoa import duplication_handler
from caffoa.base_writer import BaseWriter
from caffoa.converter import to_camelcase
from caffoa.model import ModelData, MemberData


class ModelWriter(BaseWriter):
    def __init__(self, version: int, namespace: str, output_folder: str):
        super().__init__(version)
        self.output_folder = output_folder
        self.namespace = namespace
        self.additional_imports = list()

        self.model_template = self.load_template("ModelTemplate.cs")
        self.prop_template = self.load_template("ModelPropertyTemplate.cs")

    def write(self, models: List[ModelData]):
        os.makedirs(self.output_folder, exist_ok=True)
        dates_in_models = False
        for model in models:
            if duplication_handler.should_generate(model.name):
                self.write_model(model)
                duplication_handler.store_generated(model.name)
                if self.dates_in_model(model):
                    dates_in_models = True
        if dates_in_models:
            self.write_custom_date_converter()

    @staticmethod
    def dates_in_model(model: ModelData):
        for prop in model.properties:
            if prop.is_date:
                return True
        return False

    def write_model(self, model: ModelData):
        imports = [f"using {key};\n" for key in model.imports]
        imports.extend([f"using {key};\n" for key in self.additional_imports])
        # remove duplicates but keep order:
        imports = list(dict.fromkeys(imports))
        parent = f" : {model.parent}" if model.parent else ""
        formatted_properties = []
        for prop in model.properties:
            formatted_properties.append(self.format_property(prop))
        formatted_updater = self.format_updater(model)

        file_name = os.path.abspath(self.output_folder + f"/{model.name}.generated.cs")
        logging.info(f"Writing class {model.name} -> {file_name}")

        description = "/// AUTOGENERED BY caffoa ///\n\t"
        if model.description is not None:
            model_description = model.description.strip().replace("\n", "\n\t/// ")
            description += f"/// <summary>\n\t/// {model_description}\n\t/// </summary>\n\t"

        with open(file_name, "w", encoding="utf-8") as f:
            f.write(self.model_template.format(NAMESPACE=self.namespace,
                                               NAME=model.name,
                                               RAWNAME=model.rawname,
                                               PROPERTIES="\n\n".join(formatted_properties),
                                               UPDATEPROPS=formatted_updater,
                                               IMPORTS="".join(imports),
                                               PARENTS=parent,
                                               DESCRIPTION=description))

    def format_property(self, property: MemberData) -> str:
        extra = ""
        if property.is_date:
            extra = "[JsonConverter(typeof(CustomJsonDateConverter))]\n\t\t"
        default_str = ""
        if property.default_value is not None:
            default_str = f" = {property.default_value};"

        json_property_extra = ""
        if property.is_required and property.nullable:
            json_property_extra = ", Required = Required.AllowNull"
        elif property.is_required:
            json_property_extra = ", Required = Required.Always"

        description = ""
        if property.description is not None:
            description = property.description.strip().replace("\n", "\n\t\t/// ")
            description = f"/// <summary>\n\t\t/// {description}\n\t\t/// </summary>\n\t\t"

        enums = ""
        if property.enums is not None:
            enums = f"{property.enums}\n\n\t\t"

        return self.prop_template.format(TYPE=property.typename,
                                         NAMELOWER=property.name,
                                         NAMEUPPER=to_camelcase(property.name),
                                         JSON_EXTRA=extra,
                                         DEFAULT=default_str,
                                         JSON_PROPERTY_EXTRA=json_property_extra,
                                         DESCRIPTION=description,
                                         ENUMS=enums)

    def format_updater(self, model: ModelData) -> str:
        props_update = []
        for prop in model.properties:
            name = to_camelcase(prop.name)
            special_call = ""
            if prop.is_list:
                if prop.is_basic_type:
                    special_call = ".ToList()"
            elif not prop.is_basic_type:
                special_call = f".To{prop.typename}()"
                if prop.nullable:
                    special_call = "?" + special_call
            props_update.append(f"{name} = other.{name}{special_call};")
        formatted = "\n\t\t\t".join(props_update)
        if model.parent:
            formatted = f"UpdateWith{model.parent}(other);\n\t\t\t{formatted}"
        return formatted

    def write_custom_date_converter(self):
        template = self.load_template("DateSerializer.cs")
        file_name = os.path.abspath(self.output_folder + f"/CustomJsonDateConverter.generated.cs")
        logging.info(f"Writing CustomJsonDateConverter -> {file_name}")
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(template.format(NAMESPACE=self.namespace))
